@@ -1,11 +1,14 @@
-"""OCFL Object Store library."""
+"""OCFL Storage Root library.
+
+This code uses PyFilesystem (import fs) exclusively for access to files. This
+should enable application beyond the operating system filesystem.
+"""
+import fs
+from fs.copy import copy_dir
 import hashlib
 import json
-import os
-import os.path
 import re
 import logging
-from shutil import copyfile, copytree
 try:
     from urllib.parse import quote_plus  # py3
 except ImportError:                      # pragma: no cover -- py2
@@ -15,37 +18,41 @@ from .digest import file_digest
 from .disposition import get_dispositor
 from .namaste import find_namastes, Namaste
 from .object import Object
+from .pyfs import open_fs, ocfl_walk, ocfl_opendir
 from .validator import Validator
 
 
 class StoreException(Exception):
-    """Exception class for OCFL Store."""
+    """Exception class for OCFL Storage Root."""
 
     pass
 
 
 class Store(object):
-    """Class for handling OCFL Object Stores."""
+    """Class for handling OCFL Storage Root and include OCFL Objects."""
 
     def __init__(self, root=None, disposition=None, lax_digests=False):
-        """Initialize OCFL Object Store."""
+        """Initialize OCFL Storage Root."""
         self.root = root
         self.disposition = disposition
         self.lax_digests = lax_digests
         self._dispositor = None
         #
         self.declaration_tvalue = 'ocfl_1.0'
+        self.spec_file = 'ocfl_1.0.txt'
+        self.layout_file = 'ocfl_layout.json'
+        #
+        self.root_fs = None
         self.num_traversal_errors = 0
+        self.extension = None
+        self.description = None
 
-    @property
-    def spec_file(self):
-        """Path of optional storage root spec file."""
-        return os.path.join(self.root, 'ocfl_1.0.txt')
-
-    @property
-    def layout_file(self):
-        """Path of optional storage root layout file."""
-        return os.path.join(self.root, 'ocfl_layout.json')
+    def open_root_fs(self, create=False):
+        """Open pyfs filesystem for this OCFL storage root."""
+        try:
+            self.root_fs = open_fs(self.root, create=create)
+        except (fs.opener.errors.OpenerError, fs.errors.CreateFailed) as e:
+            raise StoreException("Failed to OCFL storage root filesystem '%s' (%s)" % (self.root, str(e)))
 
     @property
     def dispositor(self):
@@ -58,119 +65,129 @@ class Store(object):
         return self._dispositor
 
     def traversal_error(self, message):
-        """Record error traversing storage root."""
+        """Record error traversing OCFL storage root."""
         self.num_traversal_errors += 1
         logging.error(message)
 
     def object_path(self, identifier):
-        """Path to OCFL object with given identifier."""
-        return os.path.join(self.root, self.dispositor.identifier_to_path(identifier))
+        """Path to OCFL object with given identifier relative to the OCFL storage root."""
+        return self.dispositor.identifier_to_path(identifier)
 
     def initialize(self):
-        """Initialize an object store."""
-        if os.path.exists(self.root):
-            raise StoreException("OCFL Object Store root %s already exists, aborting!" % (self.root))
-        os.makedirs(self.root)
-        logging.debug("Created root directory %s" % (self.root))
+        """Create and initialize a new OCFL storage root."""
+        (parent, root_dir) = fs.path.split(self.root)
+        parent_fs = open_fs(parent)
+        if parent_fs.exists(root_dir):
+            raise StoreException("OCFL storage root %s already exists, aborting!" % (self.root))
+        self.root_fs = parent_fs.makedir(root_dir)
+        logging.debug("Created OCFL storage root at %s" % (self.root))
         # Create root declaration
-        Namaste(d=0, content=self.declaration_tvalue).write(self.root)
+        Namaste(d=0, content=self.declaration_tvalue).write(pyfs=self.root_fs)
         # Create a layout declaration
         if self.disposition is not None:
-            with open(self.layout_file, 'w') as fh:
+            with self.root_fs.open(self.layout_file, 'w') as fh:
                 layout = {'key': self.disposition,
                           'description': "Non-standard layout from ocfl-py disposition -- FIXME"}
                 json.dump(layout, fh, sort_keys=True, indent=2)
         logging.info("Created OCFL storage root %s" % (self.root))
 
     def check_root_structure(self):
-        """Check the storage root.
+        """Check the OCFL storage root structure.
 
-        Raises StoreException if there is an error.
+        Assumed that self.root_fs filesystem is available. Raises
+        StoreException if there is an error.
         """
-        if not os.path.exists(self.root):
-            raise StoreException("Storage root %s does not exist!" % (self.root))
-        if not os.path.isdir(self.root):
-            raise StoreException("Storage root %s is not a directory" % (self.root))
         # Storage root declaration
-        namastes = find_namastes(0, self.root)
+        namastes = find_namastes(0, pyfs=self.root_fs)
         if len(namastes) == 0:
             raise StoreException("Storage root %s lacks required 0= declaration file" % (self.root))
         elif len(namastes) > 1:
             raise StoreException("Storage root %s has more than one 0= style declaration file" % (self.root))
         elif namastes[0].tvalue != self.declaration_tvalue:
             raise StoreException("Storage root %s declaration file not as expected, got %s" % (self.root, namastes[0].filename))
-        elif not namastes[0].content_ok(self.root):
+        elif not namastes[0].content_ok(pyfs=self.root_fs):
             raise StoreException("Storage root %s required declaration file %s has invalid content" % (self.root, namastes[0].filename))
         # Specification file and layout file
-        if os.path.exists(self.spec_file) and not os.path.isfile(self.spec_file):
+        if self.root_fs.exists(self.spec_file) and not self.root_fs.isfile(self.spec_file):
             raise StoreException("Storage root %s includes a specification entry that isn't a file" % (self.root))
-        if os.path.exists(self.layout_file):
-            self.parse_layout_file()
+        self.extension, self.description = self.parse_layout_file()
         # Other files are allowed...
         return True
 
     def parse_layout_file(self):
-        """Read and parse layout file.
+        """Read and parse layout file in OCFL storage root.
 
-        Returns key and description on success, otherwise raises a StorageException.
+        Returns:
+          - (extension, description) strings on success,
+          - (None, None) if there is now layout file (it is optional)
+          - otherwise raises a StoreException.
         """
-        try:
-            with open(self.layout_file) as fh:
-                layout = json.load(fh)
-            if type(layout) != dict:
-                raise StoreException("Storage root %s has layout file that isn't a JSON object" % (self.root))
-            elif 'key' not in layout or 'description' not in layout:
-                raise StoreException("Storage root %s has layout file doesn't have required key and description entries" % (self.root))
-            return layout['key'], layout['description']
-        except Exception as e:  # FIXME - more specific?
-            raise StorageException("Sorage root %s has layout file can't be read (%s)" % (self.root, str(e)))
+        if self.root_fs.exists(self.layout_file):
+            try:
+                with self.root_fs.open(self.layout_file) as fh:
+                    layout = json.load(fh)
+                if type(layout) != dict:
+                    raise StoreException("Storage root %s has layout file that isn't a JSON object" % (self.root))
+                elif ('extension' not in layout or type(layout['extension']) != str
+                        or 'description' not in layout or type(layout['description']) != str):
+                    raise StoreException("Storage root %s has layout file doesn't have required extension and description string entries" % (self.root))
+                return layout['extension'], layout['description']
+            except Exception as e:  # FIXME - more specific?
+                raise StoreException("OCFL storage root %s has layout file that can't be read (%s)" % (self.root, str(e)))
+        else:
+            return None, None
 
     def object_paths(self):
-        """Generator for object paths for every obects in the store.
+        """Generator for object paths for every obect in the OCFL storage root.
 
         Yields (dirpath) that is the path to the directory for each object
-        located.
+        located, relative to the OCFL storage root and without a preceding /.
 
         Will log any errors seen while traversing the directory tree under the
         storage root.
         """
-        for (dirpath, dirnames, filenames) in os.walk(self.root, followlinks=True):
-            # Ignore files in root
-            if dirpath != self.root:
-                if (len(dirnames) + len(filenames)) == 0:
-                    self.traversal_error("Empty directory %d" % (dirpath))
-                # Is this directory an OCFL object (or root?). Look for any 0= file.
-                zero_eqs = list(filter(lambda x: x.startswith('0='), filenames))
+        for (dirpath, dirs, files) in ocfl_walk(self.root_fs, is_storage_root=True):
+            if dirpath == '/':
+                pass  # Ignore files in storage root
+            elif (len(dirs) + len(files)) == 0:
+                self.traversal_error("Empty directory %s" % (dirpath))
+            elif len(files) == 0:
+                pass  # Just an intermediate directory
+            else:
+                # Is this directory an OCFL object? Look for any 0= file.
+                zero_eqs = [file for file in files if file.startswith('0=')]
                 if len(zero_eqs) > 1:
-                    self.traversal_error("Multiple 0= declaration files in %s, ignoring and not descending" % (dirpath))
-                    dirnames = []
+                    self.traversal_error("Multiple 0= declaration files in %s, ignoring" % (dirpath))
                 elif len(zero_eqs) == 1:
                     declaration = zero_eqs[0]
                     match = re.match(r'''0=ocfl_object_(\d+\.\d+)''', declaration)
                     if match and match.group(1) == '1.0':
-                        yield (dirpath)
+                        yield (dirpath.lstrip('/'))
                     elif match:
                         self.traversal_error("Object with unknown version %s in %s, ignoring" % (match.group(1), dirpath))
                     else:
                         self.traversal_error("Object with unrecognized declaration %s in %s, ignoring" % (declaration, dirpath))
-                    # don't descend
-                    dirnames = []
+                else:
+                    self.traversal_error("Directory %s has file but not object declaration, ignoring" % (dirpath))
 
     def list(self):
-        """List contents of storage."""
+        """List contents of this OCFL storage root."""
+        self.open_root_fs()
         self.check_root_structure()
         num_objects = 0
         for dirpath in self.object_paths():
-            # Parse inventory to extract id
-            id = Object().id_from_inventory(dirpath)
-            print("%s -- id=%s" % (os.path.relpath(dirpath, self.root), id))
-            num_objects += 1
-        # FIXME - do some stuff in here
+            with ocfl_opendir(self.root_fs, dirpath) as obj_fs:
+                # Parse inventory to extract id
+                id = Object(obj_fs=obj_fs).id_from_inventory()
+                print("%s -- id=%s" % (dirpath, id))
+                num_objects += 1
+                # FIXME - maybe do some more stuff in here
         logging.info("Found %d OCFL Objects under root %s" % (num_objects, self.root))
 
     def validate(self, validate_objects=True, show_warnings=False, show_errors=True, check_digests=True):
-        """Validate storage root and optionally all objects."""
+        """Validate OCFL storage root and optionally all objects."""
         valid = True
+        self.open_root_fs()
         try:
             self.check_root_structure()
             logging.info("Storage root structure is VALID")
@@ -184,7 +201,7 @@ class Store(object):
                 validator = Validator(check_digests=check_digests,
                                       lax_digests=self.lax_digests,
                                       show_warnings=show_warnings)
-                if validator.validate(dirpath):
+                if validator.validate(ocfl_opendir(self.root_fs, dirpath)):
                     good_objects += 1
                 else:
                     logging.info("Object at %s in INVALID" % (dirpath))
@@ -202,7 +219,7 @@ class Store(object):
             logging.info("Not checking OCFL objects")
         if self.num_traversal_errors > 0:
             valid = False
-            logging.info("Encountered %d errors traversing storage root" % (self.num_traveral_errors))
+            logging.info("Encountered %d errors traversing storage root" % (self.num_traversal_errors))
         # FIXME - do some stuff in here
         if valid:
             logging.info("Storage root %s is VALID", self.root)
@@ -212,14 +229,18 @@ class Store(object):
 
     def add(self, object_path):
         """Add pre-constructed object from object_path."""
+        self.open_root_fs()
+        self.check_root_structure()
         # Sanity check
         o = Object()
-        inventory = o.parse_inventory(object_path)
+        o.open_fs(object_path)
+        inventory = o.parse_inventory()
         identifier = inventory['id']
+        # Now copy
         path = self.object_path(identifier)
-        logging.info("Copying from %s to %s" % (object_path, path))
+        logging.info("Copying from %s to %s" % (object_path, fs.path.join(self.root, path)))
         try:
-            copytree(object_path, path)
+            copy_dir(o.obj_fs, '/', self.root_fs, path)
             logging.info("Copied")
         except Exception as e:
             logging.error("Copy failed: " + str(e))
